@@ -55,12 +55,26 @@
     prolog:deprecated//1,
     prolog:message_location//1,
     prolog:message_line_element/2)).
+% Lang, Term versions
+:- multifile
+    prolog:message//2,              % entire message
+    prolog:error_message//2,        % 1-st argument of error term
+    prolog:message_context//2,      % Context of error messages
+    prolog:message_location//2,	    % (File) location of error messages
+    prolog:deprecated//2.	    % Deprecated features
+:- '$hide'((
+    prolog:message//2,
+    prolog:error_message//2,
+    prolog:message_context//2,
+    prolog:deprecated//2,
+    prolog:message_location//2)).
 
 :- discontiguous
     prolog_message/3.
 
 :- public
-    translate_message//1.
+    translate_message//1,           % +Message (deprecated)
+    prolog:translate_message//1.    % +Message
 
 :- create_prolog_flag(message_context, [thread], []).
 
@@ -82,22 +96,37 @@
 %         output instead of producing a final newline.
 %       - at_same_line
 %         Start the messages at the same line (instead of using ~N)
+%
+%   @deprecated  Use  code  for   message    translation   should   call
+%   prolog:translate_message//1.
+
+prolog:translate_message(Term) -->
+    translate_message(Term).
+
+%!  translate_message(+Term)// is det.
+%
+%   Translate a message term into  message   lines.  This version may be
+%   called from user and library definitions for message translation.
 
 translate_message(Term) -->
+    { nonvar(Term) },
+    (   { message_lang(Lang) },
+        prolog:message(Lang, Term)
+    ;   prolog:message(Term)
+    ),
+    !.
+translate_message(Term) -->
+    { nonvar(Term) },
     translate_message2(Term),
     !.
 translate_message(Term) -->
-    { Term = error(_, _) },
+    { nonvar(Term),
+      Term = error(_, _)
+    },
     [ 'Unknown exception: ~p'-[Term] ].
 translate_message(Term) -->
     [ 'Unknown message: ~p'-[Term] ].
 
-translate_message2(Term) -->
-    {var(Term)},
-    !,
-    [ 'Unknown message: ~p'-[Term] ].
-translate_message2(Term) -->
-    prolog:message(Term).
 translate_message2(Term) -->
     prolog_message(Term).
 translate_message2(error(resource_error(stack), Context)) -->
@@ -125,12 +154,24 @@ make_message_lines([Last],  ['~w'-[Last]|T], T) :- !.
 make_message_lines([L0|LT], ['~w'-[L0],nl|T0], T) :-
     make_message_lines(LT, T0, T).
 
+%!  term_message(+Term)//
+%
+%   Deal  with  the  formal  argument    of  error(Format,  ImplDefined)
+%   exception  terms.  The  `ImplDefined`   argument    is   handled  by
+%   swi_location//2.
+
+:- public term_message//1.
 term_message(Term) -->
     {var(Term)},
     !,
     [ 'Unknown error term: ~p'-[Term] ].
 term_message(Term) -->
-    prolog:error_message(Term).
+    { message_lang(Lang) },
+    prolog:error_message(Lang, Term),
+    !.
+term_message(Term) -->
+    prolog:error_message(Term),
+    !.
 term_message(Term) -->
     iso_message(Term).
 term_message(Term) -->
@@ -221,6 +262,12 @@ permission_error(open, source_sink, alias(Alias)) -->
     [ 'No permission to reuse alias "~p": already taken'-[Alias] ].
 permission_error(tnot, non_tabled_procedure, Pred) -->
     [ 'The argument of tnot/1 is not tabled: ~p'-[Pred] ].
+permission_error(assert, procedure, Pred) -->
+    { '$pi_head'(Pred, Head),
+      predicate_property(Head, ssu)
+    },
+    [ '~p: an SSU (Head => Body) predicate cannot have normal Prolog clauses'-
+      [Pred] ].
 permission_error(Action, Type, Object) -->
     [ 'No permission to ~w ~w `~p'''-[Action, Type, Object] ].
 
@@ -444,18 +491,17 @@ swi_message(format_argument_type(Fmt, Arg)) -->
     [ 'Illegal argument to format sequence ~~~w: ~p'-[Fmt, Arg] ].
 swi_message(format(Msg)) -->
     [ 'Format error: ~w'-[Msg] ].
-swi_message(conditional_compilation_error(unterminated, Where)) -->
-    [ 'Unterminated conditional compilation from '-[] ],
-    cond_location(Where).
+swi_message(conditional_compilation_error(unterminated, File:Line)) -->
+    [ 'Unterminated conditional compilation from '-[], url(File:Line) ].
 swi_message(conditional_compilation_error(no_if, What)) -->
     [ ':- ~w without :- if'-[What] ].
 swi_message(duplicate_key(Key)) -->
     [ 'Duplicate key: ~p'-[Key] ].
 swi_message(initialization_error(failed, Goal, File:Line)) -->
     !,
-    [ '~w:~w: ~p: false'-[File, Line, Goal] ].
+    [ url(File:Line), ': ~p: false'-[Goal] ].
 swi_message(initialization_error(Error, Goal, File:Line)) -->
-    [ '~w:~w: ~p '-[File, Line, Goal] ],
+    [ url(File:Line), ': ~p '-[Goal] ],
     translate_message(Error).
 swi_message(determinism_error(PI, det, Found, property)) -->
     (   { '$pi_head'(user:PI, Head),
@@ -484,28 +530,31 @@ det_error(fail) -->
     [ ' failed'- [] ].
 
 
-cond_location(File:Line) -->
-    { file_base_name(File, Base) },
-    [ '~w:~d'-[Base, Line] ].
+%!  swi_location(+Term)// is det.
+%
+%   Print location information for error(Formal,   ImplDefined) from the
+%   ImplDefined term.
 
+:- public swi_location//1.
 swi_location(X) -->
-    { var(X)
-    },
-    !,
-    [].
+    { var(X) },
+    !.
+swi_location(Context) -->
+    { message_lang(Lang) },
+    prolog:message_location(Lang, Context),
+    !.
 swi_location(Context) -->
     prolog:message_location(Context),
     !.
 swi_location(context(Caller, _Msg)) -->
-    { ground(Caller)
-    },
+    { ground(Caller) },
     !,
     caller(Caller).
 swi_location(file(Path, Line, -1, _CharNo)) -->
     !,
-    [ '~w:~d: '-[Path, Line] ].
+    [ url(Path:Line), ': ' ].
 swi_location(file(Path, Line, LinePos, _CharNo)) -->
-    [ '~w:~d:~d: '-[Path, Line, LinePos] ].
+    [ url(Path:Line:LinePos), ': ' ].
 swi_location(stream(Stream, Line, LinePos, CharNo)) -->
     (   { is_stream(Stream),
           stream_property(Stream, file_name(File))
@@ -514,7 +563,7 @@ swi_location(stream(Stream, Line, LinePos, CharNo)) -->
     ;   [ 'Stream ~w:~d:~d '-[Stream, Line, LinePos] ]
     ).
 swi_location(autoload(File:Line)) -->
-    [ '~w:~w: '-[File, Line] ].
+    [ url(File:Line), ': ' ].
 swi_location(_) -->
     [].
 
@@ -533,11 +582,22 @@ caller(Caller) -->
     [ '~p: '-[Caller] ].
 
 
+%!  swi_extra(+Term)// is det.
+%
+%   Extract information from the  second   argument  of an error(Formal,
+%   ImplDefined) that is printed _after_ the core of the message.
+%
+%   @see swi_location//1 uses the same term   to insert context _before_
+%   the core of the message.
+
 swi_extra(X) -->
-    { var(X)
-    },
+    { var(X) },
     !,
     [].
+swi_extra(Context) -->
+    { message_lang(Lang) },
+    prolog:message_context(Lang, Context),
+    !.
 swi_extra(Context) -->
     prolog:message_context(Context).
 swi_extra(context(_, Msg)) -->
@@ -578,10 +638,50 @@ thread_context -->
                  *        NORMAL MESSAGES       *
                  *******************************/
 
+prolog_message(welcome) -->
+    [ 'Welcome to SWI-Prolog (' ],
+    prolog_message(threads),
+    prolog_message(address_bits),
+    ['version ' ],
+    prolog_message(version),
+    [ ')', nl ],
+    prolog_message(copyright),
+    [ nl ],
+    translate_message(user_versions),
+    [ nl ],
+    prolog_message(documentaton),
+    [ nl, nl ].
+prolog_message(user_versions) -->
+    (   { findall(Msg, prolog:version_msg(Msg), Msgs),
+          Msgs \== []
+        }
+    ->  [nl],
+        user_version_messages(Msgs)
+    ;   []
+    ).
+prolog_message(deprecated(Term)) -->
+    { nonvar(Term) },
+    (   { message_lang(Lang) },
+        prolog:deprecated(Lang, Term)
+    ->  []
+    ;   prolog:deprecated(Term)
+    ->  []
+    ;   deprecated(Term)
+    ).
+prolog_message(unhandled_exception(E)) -->
+    { nonvar(E) },
+    [ 'Unhandled exception: ' ],
+    (   translate_message(E)
+    ->  []
+    ;   [ '~p'-[E] ]
+    ).
+
+%!  prolog_message(+Term)//
+
 prolog_message(initialization_error(_, E, File:Line)) -->
     !,
-    [ '~w:~d: '-[File, Line],
-      'Initialization goal raised exception:', nl
+    [ url(File:Line),
+      ': Initialization goal raised exception:', nl
     ],
     translate_message(E).
 prolog_message(initialization_error(Goal, E, _)) -->
@@ -589,8 +689,8 @@ prolog_message(initialization_error(Goal, E, _)) -->
     translate_message(E).
 prolog_message(initialization_failure(_Goal, File:Line)) -->
     !,
-    [ '~w:~d: '-[File, Line],
-      'Initialization goal failed'-[]
+    [ url(File:Line),
+      ': Initialization goal failed'-[]
     ].
 prolog_message(initialization_failure(Goal, _)) -->
     [ 'Initialization goal failed: ~p'-[Goal]
@@ -604,10 +704,10 @@ prolog_message(init_goal_syntax(Error, Text)) -->
     translate_message(Error).
 prolog_message(init_goal_failed(failed, @(Goal,File:Line))) -->
     !,
-    [ '~w:~w: ~p: false'-[File, Line, Goal] ].
+    [ url(File:Line), ': ~p: false'-[Goal] ].
 prolog_message(init_goal_failed(Error, @(Goal,File:Line))) -->
     !,
-    [ '~w:~w: ~p '-[File, Line, Goal] ],
+    [ url(File:Line), ': ~p '-[Goal] ],
     translate_message(Error).
 prolog_message(init_goal_failed(failed, Text)) -->
     !,
@@ -616,12 +716,6 @@ prolog_message(init_goal_failed(Error, Text)) -->
     !,
     [ '-g ~w: '-[Text] ],
     translate_message(Error).
-prolog_message(unhandled_exception(E)) -->
-    [ 'Unhandled exception: ' ],
-    (   translate_message2(E)
-    ->  []
-    ;   [ '~p'-[E] ]
-    ).
 prolog_message(goal_failed(Context, Goal)) -->
     [ 'Goal (~w) failed: ~p'-[Context, Goal] ].
 prolog_message(no_current_module(Module)) -->
@@ -665,8 +759,6 @@ prolog_message(unknown_in_module_user) -->
       'Please use :- dynamic or limit usage of unknown to a module.', nl,
       'See https://www.swi-prolog.org/howto/database.html'
     ].
-prolog_message(deprecated(What)) -->
-    deprecated(What).
 prolog_message(untable(PI)) -->
     [ 'Reconsult: removed tabling for ~p'-[PI] ].
 
@@ -785,7 +877,7 @@ defined_definition(Message, Spec) -->
       predicate_property(M:Head, line_count(Line))
     },
     !,
-    [ nl, '~w at ~w:~d'-[Message, File,Line] ].
+    [ nl, '~w at '-[Message], url(File:Line) ].
 defined_definition(_, _) --> [].
 
 used_search([]) -->
@@ -834,7 +926,7 @@ current_definition(Proc, Prefix) -->
       predicate_property(Head, file(File)),
       predicate_property(Head, line_count(Line))
     },
-    [ '~w~w:~d'-[Prefix,File,Line], nl ].
+    [ '~w'-[Prefix], url(File:Line), nl ].
 current_definition(_, _) --> [].
 
 pi_uhead(Module:Name/Arity, Module:Head) :-
@@ -1037,14 +1129,13 @@ prolog_message(autoload(read_index(Dir))) -->
 prolog_message(autoload(disabled(Loaded))) -->
     [ 'Disabled autoloading (loaded ~D files)'-[Loaded] ].
 prolog_message(autoload(already_defined(PI, From))) -->
-    [ ansi(code, '~p', [PI]) ],
+    code(PI),
     (   { '$pi_head'(PI, Head),
           predicate_property(Head, built_in)
         }
     ->  [' is a built-in predicate']
-    ;   [ ' is already imported from module ',
-          ansi(code, '~p', [From])
-        ]
+    ;   [ ' is already imported from module ' ],
+        code(From)
     ).
 
 swi_message(autoload(Msg)) -->
@@ -1167,31 +1258,10 @@ prolog_message(copyright) -->
     [ 'SWI-Prolog comes with ABSOLUTELY NO WARRANTY. This is free software.', nl,
       'Please run ?- license. for legal details.'
     ].
-prolog_message(user_versions) -->
-    (   { findall(Msg, prolog:version_msg(Msg), Msgs),
-          Msgs \== []
-        }
-    ->  [nl],
-        user_version_messages(Msgs)
-    ;   []
-    ).
 prolog_message(documentaton) -->
     [ 'For online help and background, visit https://www.swi-prolog.org', nl,
       'For built-in help, use ?- help(Topic). or ?- apropos(Word).'
     ].
-prolog_message(welcome) -->
-    [ 'Welcome to SWI-Prolog (' ],
-    prolog_message(threads),
-    prolog_message(address_bits),
-    ['version ' ],
-    prolog_message(version),
-    [ ')', nl ],
-    prolog_message(copyright),
-    [ nl ],
-    prolog_message(user_versions),
-    [ nl ],
-    prolog_message(documentaton),
-    [ nl, nl ].
 prolog_message(about) -->
     [ 'SWI-Prolog version (' ],
     prolog_message(threads),
@@ -1243,11 +1313,11 @@ query_result(more(Bindings, Delays, Residuals)) -->
     result(Bindings, Delays, Residuals),
     prompt(more, Bindings, Delays, Residuals).
 query_result(help) -->
-    [ nl, 'Actions:'-[], nl, nl,
-      '; (n, r, space, TAB): redo    t:          trace & redo'-[], nl,
-      'b:                    break   c (a, RET): exit'-[], nl,
-      'w:                    write   p           print'-[], nl,
-      'h (?):                help'-[],
+    [ ansi(bold, '  Possible actions:', []), nl,
+      '  ; (n,r,space,TAB): redo              | t:         trace&redo'-[], nl,
+      '  *:                 show choicepoint  | c (a,RET): stop'-[], nl,
+      '  w:                 write             | p:         print'-[], nl,
+      '  b:                 break             | h (?):     help'-[],
       nl, nl
     ].
 query_result(action) -->
@@ -1459,6 +1529,11 @@ history_events([Nr/Event|T]) -->
     history_events(T).
 
 
+%!  user_version_messages(+Terms)//
+%
+%   Helper for the `welcome`  message   to  print information registered
+%   using version/1.
+
 user_version_messages([]) --> [].
 user_version_messages([H|T]) -->
     user_version_message(H),
@@ -1467,7 +1542,7 @@ user_version_messages([H|T]) -->
 %!  user_version_message(+Term)
 
 user_version_message(Term) -->
-    translate_message2(Term), !, [nl].
+    translate_message(Term), !, [nl].
 user_version_message(Atom) -->
     [ '~w'-[Atom], nl ].
 
@@ -1692,9 +1767,6 @@ prolog_message(backcomp(init_file_moved(FoundFile))) -->
 		 *          DEPRECATED		*
 		 *******************************/
 
-deprecated(Term) -->
-    prolog:deprecated(Term),
-    !.
 deprecated(set_prolog_stack(_Stack,limit)) -->
     [ 'set_prolog_stack/2: limit(Size) sets the combined limit.'-[], nl,
       'See https://www.swi-prolog.org/changes/stack-limit.html'
@@ -1719,6 +1791,79 @@ tripwire_context(_, Ctx) -->
     [ '~p'-[Ctx] ].
 
 
+		 /*******************************
+		 *     INTERNATIONALIZATION	*
+		 *******************************/
+
+:- create_prolog_flag(message_language, default, []).
+
+%!  message_lang(-Lang) is multi.
+%
+%   True when Lang is a language id  preferred for messages. Starts with
+%   the most specific language (e.g., `nl_BE`) and ends with `en`.
+
+message_lang(Lang) :-
+    current_message_lang(Lang0),
+    (   Lang0 == en
+    ->  Lang = en
+    ;   sub_atom(Lang0, 0, _, _, en_)
+    ->  longest_id(Lang0, Lang)
+    ;   (   longest_id(Lang0, Lang)
+        ;   Lang = en
+        )
+    ).
+
+longest_id(Lang, Id) :-
+    split_string(Lang, "_-", "", [H|Components]),
+    longest_prefix(Components, Taken),
+    atomic_list_concat([H|Taken], '_', Id).
+
+longest_prefix([H|T0], [H|T]) :-
+    longest_prefix(T0, T).
+longest_prefix(_, []).
+
+%!  current_message_lang(-Lang) is det.
+%
+%   Get the current language for messages.
+
+current_message_lang(Lang) :-
+    (   current_prolog_flag(message_language, Lang0),
+        Lang0 \== default
+    ->  Lang = Lang0
+    ;   os_user_lang(Lang0)
+    ->  clean_encoding(Lang0, Lang1),
+        set_prolog_flag(message_language, Lang1),
+        Lang = Lang1
+    ;   Lang = en
+    ).
+
+os_user_lang(Lang) :-
+    current_prolog_flag(windows, true),
+    win_get_user_preferred_ui_languages(name, [Lang|_]).
+os_user_lang(Lang) :-
+    catch(setlocale(messages, _, ''), _, fail),
+    setlocale(messages, Lang, Lang).
+os_user_lang(Lang) :-
+    getenv('LANG', Lang).
+
+
+clean_encoding(Lang0, Lang) :-
+    (   sub_atom(Lang0, A, _, _, '.')
+    ->  sub_atom(Lang0, 0, A, _, Lang)
+    ;   Lang = Lang0
+    ).
+
+		 /*******************************
+		 *          PRIMITIVES		*
+		 *******************************/
+
+code(Term) -->
+    code('~p', Term).
+
+code(Format, Term) -->
+    [ ansi(code, Format, [Term]) ].
+
+
 		 /*******************************
 		 *        DEFAULT THEME		*
 		 *******************************/
@@ -1768,35 +1913,37 @@ default_theme(message(Level),         Attrs) :-
 %   Print an error message using a term as generated by the exception
 %   system.
 
+print_message(Level, _Term) :-
+    msg_property(Level, stream(S)),
+    stream_property(S, error(true)),
+    !.
 print_message(Level, Term) :-
     setup_call_cleanup(
-        push_msg(Term),
-        print_message_guarded(Level, Term),
-        pop_msg),
+        push_msg(Term, Stack),
+        ignore(print_message_guarded(Level, Term)),
+        pop_msg(Stack)),
     !.
 print_message(Level, Term) :-
     (   Level \== silent
-    ->  format(user_error, 'Recursive ~w message: ~q~n', [Level, Term])
+    ->  format(user_error, 'Recursive ~w message: ~q~n', [Level, Term]),
+        backtrace(20)
     ;   true
     ).
 
-push_msg(Term) :-
+push_msg(Term, Messages) :-
     nb_current('$inprint_message', Messages),
     !,
     \+ ( '$member'(Msg, Messages),
          Msg =@= Term
        ),
-    b_setval('$inprint_message', [Term|Messages]).
-push_msg(Term) :-
+    Stack = [Term|Messages],
+    b_setval('$inprint_message', Stack).
+push_msg(Term, []) :-
     b_setval('$inprint_message', [Term]).
 
-pop_msg :-
-    (   nb_current('$inprint_message', [_|Messages]),
-        Messages \== []
-    ->  b_setval('$inprint_message', Messages)
-    ;   nb_delete('$inprint_message'),              % delete history
-        b_setval('$inprint_message', [])
-    ).
+pop_msg(Stack) :-
+    nb_delete('$inprint_message'),              % delete history
+    b_setval('$inprint_message', Stack).
 
 print_message_guarded(Level, Term) :-
     (   must_print(Level, Term)
@@ -1847,13 +1994,13 @@ print_system_message(Term, Kind, Lines) :-
     Term \= error(syntax_error(_), _),
     msg_property(Kind, location_prefix(File:Line, LocPrefix, LinePrefix)),
     !,
+    to_list(LocPrefix, LocPrefixL),
     insert_prefix(Lines, LinePrefix, Ctx, PrefixLines),
-    '$append'([ begin(Kind, Ctx),
-                LocPrefix,
-                nl
-              | PrefixLines
-              ],
-              [ end(Ctx)
+    '$append'([ [begin(Kind, Ctx)],
+                LocPrefixL,
+                [nl],
+                PrefixLines,
+                [end(Ctx)]
               ],
               AllLines),
     msg_property(Kind, stream(Stream)),
@@ -1869,6 +2016,12 @@ print_system_message(_, Kind, Lines) :-
     msg_property(Kind, stream(Stream)),
     print_message_lines(Stream, kind(Kind), Lines).
 
+to_list(ListIn, List) :-
+    is_list(ListIn),
+    !,
+    List = ListIn.
+to_list(NonList, [NonList]).
+
 :- multifile
     user:message_property/2.
 
@@ -1881,22 +2034,25 @@ msg_property(Kind, prefix(Prefix)) :-
 msg_property(_, prefix('~N')) :- !.
 msg_property(query, stream(user_output)) :- !.
 msg_property(_, stream(user_error)) :- !.
-msg_property(error,
+msg_property(error, tag('ERROR')).
+msg_property(warning, tag('Warning')).
+msg_property(Level,
              location_prefix(File:Line,
-                             '~NERROR: ~w:~d:'-[File,Line],
-                             '~NERROR:    ')) :- !.
-msg_property(warning,
-             location_prefix(File:Line,
-                             '~NWarning: ~w:~d:'-[File,Line],
-                             '~NWarning:    ')) :- !.
+                             ['~N~w: '-[Tag], url(File:Line), ':'],
+                             '~N~w:    '-[Tag])) :-
+    include_msg_location(Level),
+    msg_property(Level, tag(Tag)).
 msg_property(error,   wait(0.1)) :- !.
 
+include_msg_location(warning).
+include_msg_location(error).
+
 msg_prefix(debug(_), Prefix) :-
     msg_context('~N% ', Prefix).
-msg_prefix(warning, Prefix) :-
-    msg_context('~NWarning: ', Prefix).
-msg_prefix(error, Prefix) :-
-    msg_context('~NERROR: ', Prefix).
+msg_prefix(Level, Prefix) :-
+    msg_property(Level, tag(Tag)),
+    atomics_to_string(['~N', Tag, ': '], Prefix0),
+    msg_context(Prefix0, Prefix).
 msg_prefix(informational, '~N% ').
 msg_prefix(information,   '~N% ').
 
@@ -2028,11 +2184,29 @@ line_element(S, ansi(_, Fmt, Args)) :-
 line_element(S, ansi(_, Fmt, Args, _Ctx)) :-
     !,
     safe_format(S, Fmt, Args).
+line_element(S, url(URL)) :-
+    !,
+    print_link(S, URL).
+line_element(S, url(_URL, Fmt-Args)) :-
+    !,
+    safe_format(S, Fmt, Args).
+line_element(S, url(_URL, Fmt)) :-
+    !,
+    safe_format(S, Fmt, []).
 line_element(_, begin(_Level, _Ctx)) :- !.
 line_element(_, end(_Ctx)) :- !.
 line_element(S, Fmt) :-
     safe_format(S, Fmt, []).
 
+print_link(S, File:Line:Column) :-
+    !,
+    safe_format(S, '~w:~d:~d', [File, Line, Column]).
+print_link(S, File:Line) :-
+    !,
+    safe_format(S, '~w:~d', [File, Line]).
+print_link(S, File) :-
+    safe_format(S, '~w', [File]).
+
 %!  safe_format(+Stream, +Format, +Args) is det.
 
 safe_format(S, Fmt, Args) :-
@@ -2041,7 +2215,7 @@ safe_format(S, Fmt, Args) :-
           format_failed(S,Fmt,Args,E)).
 
 format_failed(S, _Fmt, _Args, E) :-
-    E = error(io_error(_,S),_),
+    stream_property(S, error(true)),
     !,
     throw(E).
 format_failed(S, Fmt, Args, error(E,_)) :-
@@ -2077,6 +2251,14 @@ actions_to_format([ansi(_Attrs, Fmt0, Args0)|Tail], Fmt, Args) :-
     actions_to_format(Tail, Fmt1, Args1),
     atom_concat(Fmt0, Fmt1, Fmt),
     append_args(Args0, Args1, Args).
+actions_to_format([url(Pos)|Tail], Fmt, Args) :-
+    !,
+    actions_to_format(Tail, Fmt1, Args1),
+    url_actions_to_format(url(Pos), Fmt1, Args1, Fmt, Args).
+actions_to_format([url(URL, Label)|Tail], Fmt, Args) :-
+    !,
+    actions_to_format(Tail, Fmt1, Args1),
+    url_actions_to_format(url(URL, Label), Fmt1, Args1, Fmt, Args).
 actions_to_format([Fmt0-Args0|Tail], Fmt, Args) :-
     !,
     actions_to_format(Tail, Fmt1, Args1),
@@ -2101,14 +2283,33 @@ action_skip(flush).
 action_skip(begin(_Level, _Ctx)).
 action_skip(end(_Ctx)).
 
+url_actions_to_format(url(File:Line:Column), Fmt1, Args1, Fmt, Args) :-
+    !,
+    atom_concat('~w:~d:~d', Fmt1, Fmt),
+    append_args([File,Line,Column], Args1, Args).
+url_actions_to_format(url(File:Line), Fmt1, Args1, Fmt, Args) :-
+    !,
+    atom_concat('~w:~d', Fmt1, Fmt),
+    append_args([File,Line], Args1, Args).
+url_actions_to_format(url(File), Fmt1, Args1, Fmt, Args) :-
+    !,
+    atom_concat('~w', Fmt1, Fmt),
+    append_args([File], Args1, Args).
+url_actions_to_format(url(_URL, Label), Fmt1, Args1, Fmt, Args) :-
+    !,
+    atom_concat('~w', Fmt1, Fmt),
+    append_args([Label], Args1, Args).
+
+
 append_args(M:Args0, Args1, M:Args) :-
     !,
     strip_module(Args1, _, A1),
-    '$append'(Args0, A1, Args).
+    to_list(Args0, Args01),
+    '$append'(Args01, A1, Args).
 append_args(Args0, Args1, Args) :-
     strip_module(Args1, _, A1),
-    '$append'(Args0, A1, Args).
-
+    to_list(Args0, Args01),
+    '$append'(Args01, A1, Args).
 
                  /*******************************
                  *    MESSAGES TO PRINT ONCE    *
