@@ -41,7 +41,9 @@
             doc_browser/1               % +What
           ]).
 :- use_module(library(pldoc)).
+:- if(exists_source(library(http/thread_httpd))).
 :- use_module(library(http/thread_httpd)).
+:- endif.
 :- use_module(library(http/http_parameters)).
 :- use_module(library(http/html_write)).
 :- use_module(library(http/mimetype)).
@@ -68,6 +70,7 @@
 :- use_module(pldoc(doc_util)).
 :- use_module(pldoc(doc_access)).
 :- use_module(pldoc(doc_pack)).
+:- use_module(pldoc(man_index)).
 
 /** <module> Documentation server
 
@@ -104,7 +107,7 @@ doc_enable(false) :-
 %!  doc_server(?Port, +Options) is det.
 %
 %   Start a documentation server in the  current Prolog process. The
-%   server is started in a seperate   thread.  Options are handed to
+%   server is started in a separate   thread.  Options are handed to
 %   http_server/2.  In  addition,   the    following   options   are
 %   recognised:
 %
@@ -144,6 +147,7 @@ doc_server(Port, _) :-
     doc_enable(true),
     catch(doc_current_server(Port), _, fail),
     !.
+:- if(current_predicate(http_server/2)).
 doc_server(Port, Options) :-
     doc_enable(true),
     prepare_editor,
@@ -155,6 +159,7 @@ doc_server(Port, Options) :-
                   ], HTTPOptions),
     http_server(http_dispatch, HTTPOptions),
     assertz(doc_server_port(Port)).
+:- endif.
 
 %!  doc_current_server(-Port) is det.
 %
@@ -175,6 +180,10 @@ doc_current_server(Port) :-
     ;   existence_error(http_server, pldoc)
     ).
 
+:- if(\+current_predicate(http_current_server/2)).
+http_current_server(_,_) :- fail.
+:- endif.
+
 %!  doc_browser is det.
 %!  doc_browser(+What) is semidet.
 %
@@ -206,7 +215,7 @@ browser_url(Name//Arity, URL) :-
 browser_url(Name/Arity, URL) :-
     !,
     must_be(atom, Name),
-    (   predicate(Name, Arity, _, _, _)
+    (   man_object_property(Name/Arity, summary(_))
     ->  format(string(S), '~q/~w', [Name, Arity]),
         http_link_to_id(pldoc_man, [predicate=S], URL)
     ;   browser_url(_:Name/Arity, URL)
@@ -713,7 +722,10 @@ pldoc_object(Request) :-
                     [ object(Atom, []),
                       header(Header, [default(true)])
                     ]),
-    atom_to_term(Atom, Obj, _),
+    (   catch(atom_to_term(Atom, Obj, _), error(_,_), fail)
+    ->  true
+    ;   atom_to_object(Atom, Obj)
+    ),
     (   prolog:doc_object_title(Obj, Title)
     ->  true
     ;   Title = Atom
